# -*- coding: utf-8 -*-

"""
todo: doc string
"""

import typing as T
import dataclasses

from func_args import NOTHING, resolve_kwargs
from boto_session_manager import BotoSesManager

from acore_server_metadata.api import Server as ServerMetadata
from acore_server_config.api import ConfigLoader

from .server import Server


@dataclasses.dataclass
class Fleet:  # pragma: no cover
    """
    Fleet of server for a given environment.

    It is just a dictionary containing all servers' data. key is ``server_id``
    value is the :class:`Server` object.

    Usage:

    .. code-block:: python

        >>> fleet = Fleet.get(bsm=..., env_name="sbx", ...)
        >>> server = fleet.get_server(server_id="sbx-blue")
    """

    servers: T.Dict[str, Server] = dataclasses.field(init=False)

    @classmethod
    def get(
        cls,
        bsm: BotoSesManager,
        env_name: str,
        parameter_name_prefix: T.Optional[str] = NOTHING,
        s3folder_config: T.Optional[str] = NOTHING,
    ):
        """
        Load all servers' data for a given environment efficiently.
        """
        # get acore_server_config
        config_loader = ConfigLoader.new(
            **resolve_kwargs(
                env_name=env_name,
                parameter_name_prefix=parameter_name_prefix,
                s3folder_config=s3folder_config,
                bsm=bsm,
            )
        )
        server_id_list = [server.id for _, server in config_loader.iter_servers()]
        # get acore_server_metadata
        server_metadata_mapper = ServerMetadata.batch_get_server(
            ids=server_id_list,
            ec2_client=bsm.ec2_client,
            rds_client=bsm.rds_client,
        )
        # put them together
        servers = dict()
        for server_id in server_id_list:
            server_name = server_id.split("-", 1)[1]
            metadata = server_metadata_mapper.get(server_id)
            if metadata is None:
                metadata = ServerMetadata(id=server_id)
            server = Server(
                config=config_loader.get_server(server_name=server_name),
                metadata=metadata,
            )
            servers[server_id] = server

        fleet = cls()
        fleet.servers = servers
        return fleet

    def get_server(self, server_id: str) -> Server:
        """
        Get a server by its id.
        """
        return self.servers[server_id]



# -*- coding: utf-8 -*-

"""
todo: doctring
"""

import dataclasses
from functools import cached_property

from s3pathlib import S3Path
from boto_session_manager import BotoSesManager

from .wserver_infra_exports import StackExports
from .server import Server
from .fleet import Fleet


@dataclasses.dataclass
class Manager:
    """
    This manager class is to simplify executing Server workflow methods.
    """

    aws_profile: str = dataclasses.field()
    env_name: str = dataclasses.field()

    @cached_property
    def bsm(self) -> "BotoSesManager":
        return BotoSesManager(profile_name=self.aws_profile)

    @cached_property
    def s3dir_env_workflow(self) -> S3Path:
        return S3Path(
            f"s3://{self.bsm.aws_account_alias}-{self.bsm.aws_region}-data"
            f"/projects/acore_server/workflows/{self.env_name}/"
        ).to_dir()

    @cached_property
    def stack_exports(self) -> StackExports:
        stack_exports = StackExports(env_name=self.env_name)
        stack_exports.load(cf_client=self.bsm.cloudformation_client)
        return stack_exports

    @cached_property
    def fleet(self) -> Fleet:
        return Fleet.get(bsm=self.bsm, env_name=self.env_name)

    @property
    def blue(self) -> Server:
        return self.fleet.get_server(server_id=f"{self.env_name}-blue")

    @property
    def green(self) -> Server:
        return self.fleet.get_server(server_id=f"{self.env_name}-green")

    @property
    def black(self) -> Server:
        return self.fleet.get_server(server_id=f"{self.env_name}-black")

    @property
    def white(self) -> Server:
        return self.fleet.get_server(server_id=f"{self.env_name}-white")

    @property
    def yellow(self) -> Server:
        return self.fleet.get_server(server_id=f"{self.env_name}-yellow")

    @property
    def orange(self) -> Server:
        return self.fleet.get_server(server_id=f"{self.env_name}-orange")



# -*- coding: utf-8 -*-

from datetime import datetime, timezone


def get_utc_now() -> datetime:
    return datetime.utcnow().replace(tzinfo=timezone.utc)


def prompt_for_confirm(msg: str):
    entered = input(f"{msg} Enter 'YES' to proceed: ")
    if entered != "YES":
        raise KeyboardInterrupt()



# -*- coding: utf-8 -*-

from acore_server import api


def test():
    _ = api
    _ = api.Server
    _ = api.Fleet
    _ = api.InfraStackExports

    _ = api.Server
    _ = api.Server.id
    _ = api.Server.env_name
    _ = api.Server.server_name
    _ = api.Server.create_ec2
    _ = api.Server.create_rds_from_scratch
    _ = api.Server.create_rds_from_snapshot
    _ = api.Server.start_ec2
    _ = api.Server.start_rds
    _ = api.Server.associate_eip_address
    _ = api.Server.update_db_master_password
    _ = api.Server.stop_ec2
    _ = api.Server.stop_rds
    _ = api.Server.delete_ec2
    _ = api.Server.delete_rds
    _ = api.Server.bootstrap
    _ = api.Server.run_check_server_status_cron_job
    _ = api.Server.stop_check_server_status_cron_job
    _ = api.Server.run_game_server
    _ = api.Server.stop_game_server
    _ = api.Server.wow_status
    _ = api.Server.create_ssh_tunnel
    _ = api.Server.list_ssh_tunnel


if __name__ == "__main__":
    from acore_server.tests import run_cov_test

    run_cov_test(__file__, "acore_server.api", preview=False)



# -*- coding: utf-8 -*-

"""
todo: doc string
"""

import typing as T
import json
import dataclasses
from boto_session_manager import BotoSesManager
from aws_console_url.api import AWSConsole

try:
    from rich import print as rprint
except ImportError:
    pass

from .logger import logger

aws_console = AWSConsole(aws_region="us-east-1")

if T.TYPE_CHECKING:  # pragma: no cover
    from .server import Server


class ServerHelperMixin:  # pragma: no cover
    """
    Server Helper Mixin class that contains all the server helper methods.
    """

    @logger.emoji_block(
        msg="🔍🛠Show server config",
        emoji="🔍",
    )
    def show_server_config(
        self: "Server",
    ):
        logger.info(json.dumps(dataclasses.asdict(self.config), indent=4))

    @logger.emoji_block(
        msg="🔍🖥🛢Show server status",
        emoji="🔍",
    )
    def show_server_status(
        self: "Server",
    ):
        logger.info(f"🖥is_ec2_exists: {self.metadata.is_ec2_exists()}")
        logger.info(f"🛢is_rds_exists: {self.metadata.is_rds_exists()}")
        logger.info(f"🖥is_ec2_running: {self.metadata.is_ec2_running()}")
        logger.info(f"🛢is_rds_running: {self.metadata.is_rds_running()}")

    @logger.emoji_block(
        msg="🔍🛠Show server config",
        emoji="🔍",
    )
    def show_aws_link(
        self: "Server",
    ):
        if self.metadata.is_ec2_exists():
            url = aws_console.ec2.get_instance(self.metadata.ec2_inst.id)
            logger.info(f"🌐🖥preview EC2: {url}")
        else:
            logger.info("🛑EC2 does not exist")
        if self.metadata.is_rds_exists():
            url = aws_console.rds.get_database_instance(self.metadata.rds_inst.id)
            logger.info(f"🌐🖥preview RDS: {url}")
        else:
            logger.info("🛑RDS does not exist!")



# -*- coding: utf-8 -*-

from vislog import VisLog

# I am trying to use rich for logging, I intentionally this code here
# I haven't made decision using it right now
try:  # pragma: no cover
    from rich import print as rprint

    class RichLogger:
        def info(self, msg: str):
            rprint(msg)

    _logger = RichLogger()
except ImportError:  # pragma: no cover
    _logger = None

_logger = None

logger = VisLog(
    name="acore_server",
    logger=_logger,
    log_format="%(message)s",
)



# -*- coding: utf-8 -*-

"""
todo: doc string
"""

import typing as T
import dataclasses

from func_args import NOTHING, resolve_kwargs
from boto_session_manager import BotoSesManager
from simple_aws_ec2.api import Ec2Instance
from simple_aws_rds.api import RDSDBInstance

from acore_server_metadata.api import Server as ServerMetadata
from acore_server_config.api import (
    Server as ServerConfig,
    Ec2ConfigLoader,
    ConfigLoader,
)

from .exc import (
    ServerNotFoundError,
    ServerAlreadyExistsError,
    FailedToStartServerError,
    FailedToStopServerError,
)
from .server_operation_mixin import ServerOperationMixin
from .server_workflow_mixin import ServerWorkflowMixin
from .server_helper_mixin import ServerHelperMixin


@dataclasses.dataclass
class Server(
    ServerOperationMixin,
    ServerWorkflowMixin,
    ServerHelperMixin,
):  # pragma: no cover
    """
    A data container that holds both the config and metadata of a server.

    Usage:

    .. code-block:: python

        # if you are on develop's laptop or lambda
        >>> server = Server.get(bsm=..., server_id="sbx-blue", ...)
        # if you are on game server EC2
        >>> server = Server.from_ec2_inside()

    Server operation methods:

    - :meth:`Server.run_ec2 <acore_server.server_operation_mixin.ServerOperationMixin.run_ec2`
    - :meth:`Server.run_rds <acore_server.server_operation_mixin.ServerOperationMixin.run_rds`
    - :meth:`Server.start_ec2 <acore_server.server_operation_mixin.ServerOperationMixin.start_ec2`
    - :meth:`Server.start_rds <acore_server.server_operation_mixin.ServerOperationMixin.start_rds`
    - :meth:`Server.associate_eip_address <acore_server.server_operation_mixin.ServerOperationMixin.associate_eip_address`
    - :meth:`Server.update_db_master_password <acore_server.server_operation_mixin.ServerOperationMixin.update_db_master_password`
    - :meth:`Server.stop_ec2 <acore_server.server_operation_mixin.ServerOperationMixin.stop_ec2`
    - :meth:`Server.stop_rds <acore_server.server_operation_mixin.ServerOperationMixin.stop_rds`
    - :meth:`Server.delete_ec2 <acore_server.server_operation_mixin.ServerOperationMixin.delete_ec2`
    - :meth:`Server.delete_rds <acore_server.server_operation_mixin.ServerOperationMixin.delete_rds`
    - :meth:`Server.bootstrap <acore_server.server_operation_mixin.ServerOperationMixin.bootstrap`
    - :meth:`Server.run_server <acore_server.server_operation_mixin.ServerOperationMixin.run_server`
    - :meth:`Server.stop_server <acore_server.server_operation_mixin.ServerOperationMixin.stop_server`
    - :meth:`Server.count_online_player <acore_server.server_operation_mixin.ServerOperationMixin.count_online_player`
    - :meth:`Server.create_ssh_tunnel <acore_server.server_operation_mixin.ServerOperationMixin.create_ssh_tunnel`
    - :meth:`Server.list_ssh_tunnel <acore_server.server_operation_mixin.ServerOperationMixin.list_ssh_tunnel`
    - :meth:`Server.test_ssh_tunnel <acore_server.server_operation_mixin.ServerOperationMixin.test_ssh_tunnel`
    - :meth:`Server.kill_ssh_tunnel <acore_server.server_operation_mixin.ServerOperationMixin.kill_ssh_tunnel`

    :param config: See https://acore-server-config.readthedocs.io/en/latest/acore_server_config/config/define/server.html#acore_server_config.config.define.server.Server
    :param metadata: See https://github.com/MacHu-GWU/acore_server_metadata-project/blob/main/acore_server_metadata/server/server.py
    """

    config: ServerConfig
    metadata: ServerMetadata

    @classmethod
    def get(
        cls,
        bsm: BotoSesManager,
        server_id: str,
        parameter_name_prefix: T.Optional[str] = NOTHING,
        s3folder_config: T.Optional[str] = NOTHING,
    ):
        """
        指定一个 server_id, 读取它的 config 和 metadata, 然后返回一个 Server 对象.

        :param bsm: BotoSesManager 对象
        :param server_id: server id, the naming convention is ``${env_name}-${server_name}``
        :param parameter_name_prefix: see
            https://acore-server-config.readthedocs.io/en/latest/acore_server_config/config/loader.html#acore_server_config.config.loader.get_config
        :param s3folder_config: see
            https://acore-server-config.readthedocs.io/en/latest/acore_server_config/config/loader.html#acore_server_config.config.loader.get_config
        """
        env_name, server_name = server_id.split("-", 1)
        # get acore_server_config
        config_loader = ConfigLoader.new(
            **resolve_kwargs(
                env_name=env_name,
                parameter_name_prefix=parameter_name_prefix,
                s3folder_config=s3folder_config,
                bsm=bsm,
            )
        )
        server_config = config_loader.get_server(server_name=server_name)
        # get acore_server_metadata
        server_metadata = ServerMetadata.get_server(
            id=server_id,
            ec2_client=bsm.ec2_client,
            rds_client=bsm.rds_client,
        )
        # put them together
        server = cls(
            config=server_config,
            metadata=server_metadata,
        )
        return server

    @classmethod
    def from_ec2_inside(cls):
        """
        用 "自省" 的方式从 EC2 instance 里面读取 server_id, 读取 config 和 metadata,
        然后返回一个 Server 对象.
        """
        server_metadata = ServerMetadata.from_ec2_inside()
        server_config = Ec2ConfigLoader.load(server_id=server_metadata.id)
        # put them together
        server = cls(
            config=server_config,
            metadata=server_metadata,
        )
        return server

    @property
    def id(self) -> str:
        """
        Server id, the naming convention is ``${env_name}-${server_name}``
        """
        return self.metadata.id

    @property
    def env_name(self) -> str:
        """
        Environment name, e.g. ``sbx``, ``tst``, ``prd``.
        """
        return self.metadata.env_name

    @property
    def server_name(self) -> str:
        """
        Server name, e.g. ``blue``, ``green``.
        """
        return self.metadata.server_name

    def ensure_ec2_not_exists(
        self: "Server",
        bsm: "BotoSesManager",
    ) -> T.Optional[Ec2Instance]:
        ec2_inst = self.metadata.get_ec2(bsm.ec2_client, id=self.id)
        if ec2_inst is not None:  # pragma: no cover
            raise ServerAlreadyExistsError(f"EC2 instance {self.id!r} already exists!")
        return ec2_inst

    def ensure_ec2_exists(
        self: "Server",
        bsm: "BotoSesManager",
    ) -> T.Optional[Ec2Instance]:
        ec2_inst = self.metadata.get_ec2(bsm.ec2_client, id=self.id)
        if ec2_inst is None:  # pragma: no cover
            raise ServerNotFoundError(f"EC2 instance {self.id!r} not found!")
        return ec2_inst

    def ensure_rds_not_exists(
        self: "Server",
        bsm: "BotoSesManager",
    ) -> T.Optional[RDSDBInstance]:
        rds_inst = self.metadata.get_rds(bsm.rds_client, id=self.id)
        if rds_inst is not None:  # pragma: no cover
            raise ServerAlreadyExistsError(
                f"RDS DB instance {self.id!r} already exists!"
            )
        return rds_inst

    def ensure_rds_exists(
        self: "Server",
        bsm: "BotoSesManager",
    ) -> T.Optional[RDSDBInstance]:
        rds_inst = self.metadata.get_rds(bsm.rds_client, id=self.id)
        if rds_inst is None:  # pragma: no cover
            raise ServerNotFoundError(f"RDS DB instance {self.id!r} not found!")
        return rds_inst

    def ensure_ec2_is_ready_to_start(
        self: "Server",
        bsm: "BotoSesManager",
    ) -> T.Optional[Ec2Instance]:
        ec2_inst = self.ensure_ec2_exists(bsm)
        if ec2_inst.is_ready_to_start() is False:
            raise FailedToStartServerError(
                f"EC2 instance {self.id!r} is not ready to start!"
            )
        return ec2_inst

    def ensure_ec2_is_ready_to_stop(
        self: "Server",
        bsm: "BotoSesManager",
    ) -> T.Optional[Ec2Instance]:
        ec2_inst = self.ensure_ec2_exists(bsm)
        if ec2_inst.is_ready_to_stop() is False:
            raise FailedToStopServerError(
                f"EC2 instance {self.id!r} is not ready to start!"
            )
        return ec2_inst

    def ensure_rds_is_ready_to_start(
        self: "Server",
        bsm: "BotoSesManager",
    ) -> T.Optional[RDSDBInstance]:
        rds_inst = self.ensure_rds_exists(bsm)
        if rds_inst.is_ready_to_start() is False:
            raise FailedToStopServerError(
                f"RDS instance {self.id!r} is not ready to start!"
            )
        return rds_inst

    def ensure_rds_is_ready_to_stop(
        self: "Server",
        bsm: "BotoSesManager",
    ) -> T.Optional[RDSDBInstance]:
        rds_inst = self.ensure_rds_exists(bsm)
        if rds_inst.is_ready_to_stop() is False:
            raise FailedToStopServerError(
                f"RDS instance {self.id!r} is not ready to stop!"
            )
        return rds_inst



# -*- coding: utf-8 -*-

"""
This module provides an interface for external code to read cloudformation output.
via export name. You can just copy and paste this file into the external codebase.
"""

import typing as T
import dataclasses

project_name = "wserver_infra"  # this value has to match the ``config.project_name``
project_name_snake = project_name.replace("-", "_")
project_name_slug = project_name.replace("_", "-")


@dataclasses.dataclass
class StackExports:
    """
    Read cloudformation output values from ``wserver-infra-{self.env_name}`` stack.

    Usage:

    .. code-block:: python

        >>> import boto3
        >>> cf_client = boto3.client("cloudformation")
        >>> stack_exports = StackExports(env_name="sbx")
        >>> stack_exports.load(cf_client)
        >>> stack_exports.get_db_subnet_group_name()
        ...
        >>> stack_exports.get_ssh_sg_id()
        ...
        >>> stack_exports.get_default_sg_id(server_id="sbx-blue")
        ...
        >>> stack_exports.get_ec2_sg_id(server_id="sbx-blue")
        ...
        >>> stack_exports.get_ec2_iam_role_arn()
        ...
        >>> stack_exports.get_ec2_instance_profile_arn()
        ...

    :param env_name: environment name, sbx, tst, prd, etc ...
    :param _outputs: internal cache, don't use it directly.
    :param _exports: internal cache, don't use it directly.
    """

    env_name: str = dataclasses.field()
    _outputs: T.Dict[str, str] = dataclasses.field(default_factory=dict)
    _exports: T.Dict[str, str] = dataclasses.field(default_factory=dict)

    @property
    def prefix_name_snake(self) -> str:  # pragma: no cover
        return f"{project_name_snake}-{self.env_name}"

    @property
    def prefix_name_slug(self) -> str:  # pragma: no cover
        return f"{project_name_slug}-{self.env_name}"

    @property
    def stack_name(self) -> str:
        return self.prefix_name_slug

    def load(self, cf_client):
        res = cf_client.describe_stacks(StackName=self.stack_name)
        output_list = res["Stacks"][0]["Outputs"]
        outputs = {dct["OutputKey"]: dct["OutputValue"] for dct in output_list}
        exports = {dct["ExportName"]: dct["OutputValue"] for dct in output_list}
        self._outputs = outputs
        self._exports = exports

    # --------------------------------------------------------------------------
    # access output value via export name
    # --------------------------------------------------------------------------
    def get_db_subnet_group_name(self) -> str:
        return self._exports[
            f"{self.prefix_name_slug}-db-subnet-group-name-for-private-db"
        ]

    def get_ssh_sg_id(self) -> str:
        return self._exports[f"{self.prefix_name_slug}-sg-ssh-id"]

    def get_default_sg_id(self, server_id: str) -> str:
        return self._exports[f"{self.prefix_name_slug}-{server_id}-sg-default-id"]

    def get_ec2_sg_id(self, server_id: str) -> str:
        return self._exports[f"{self.prefix_name_slug}-{server_id}-sg-ec2-id"]

    def get_ec2_iam_role_arn(self) -> str:
        return self._exports[f"{self.prefix_name_slug}-ec2-iam-role-arn"]

    def get_ec2_instance_profile_arn(self) -> str:
        return self._exports[f"{self.prefix_name_slug}-instance-profile-arn"]



# -*- coding: utf-8 -*-


class ServerNotFoundError(Exception):
    """
    Raises when a :class:`~acore_server_metadata.server.Server` is not found.
    """

    pass


class ServerNotUniqueError(Exception):
    """
    Raises when there are multiple  :class:`~acore_server_metadata.server.Server`
    has the same id.
    """

    pass


class ServerAlreadyExistsError(Exception):
    """
    Raises when try to launch a new EC2 or DB instance when there is already
    a existing one.
    """


class FailedToStartServerError(Exception):
    """
    Raises when the current EC2 and RDS state is not ready for starting.
    (server has to exist first)
    """


class FailedToStopServerError(Exception):
    """
    Raises when the current EC2 and RDS state is not ready for stopping.
    (server has to exist first)
    """



# -*- coding: utf-8 -*-

"""
todo: doc string
"""

import typing as T
import dataclasses
import json
from datetime import datetime, timezone

from boto_session_manager import BotoSesManager
from s3pathlib import S3Path
from aws_console_url.api import AWSConsole
import simple_aws_ec2.api as simple_aws_ec2
import simple_aws_rds.api as simple_aws_rds
from .exc import (
    FailedToStartServerError,
    FailedToStopServerError,
)
from .logger import logger
from .utils import get_utc_now, prompt_for_confirm
from .wserver_infra_exports import StackExports


if T.TYPE_CHECKING:  # pragma: no cover
    from .server import Server


@dataclasses.dataclass
class Workflow:
    """
    因为一个 workflow 在执行过程中会有多个步骤, 有些步骤可能会失败. 为了能在执行的过程中记录下
    执行的状态, 以便能在失败重试的时候从上一个成功的地方继续执行, 我们需要一个 Workflow 类来
    记录这些状态. 我选择将 Workflow 的数据保存在 S3 上.
    """

    workflow_id: str = dataclasses.field()

    def dump(self, bsm: BotoSesManager, s3_path: S3Path):
        s3_path.write_text(
            json.dumps(dataclasses.asdict(self)),
            bsm=bsm,
            content_type="application/json",
        )

    @classmethod
    def load(cls, bsm: BotoSesManager, s3_path: S3Path, workflow_id: str):
        if s3_path.exists(bsm=bsm) is False:
            workflow = cls(workflow_id=workflow_id)
            workflow.dump(bsm=bsm, s3_path=s3_path)
        else:
            workflow = cls(**json.loads(s3_path.read_text(bsm=bsm)))
            if workflow.workflow_id != workflow_id:
                raise ValueError(
                    f"workflow_id mismatch: {workflow.workflow_id} != {workflow_id}"
                )
        return workflow


@dataclasses.dataclass
class CreateClonedServerWorkflow(Workflow):
    ec2_ami_id: T.Optional[str] = dataclasses.field(default=None)
    db_snapshot_id: T.Optional[str] = dataclasses.field(default=None)
    ec2_inst_id: T.Optional[str] = dataclasses.field(default=None)
    db_inst_id: T.Optional[str] = dataclasses.field(default=None)
    is_ec2_ami_deleted: bool = dataclasses.field(default=False)
    is_db_snapshot_deleted: bool = dataclasses.field(default=False)

    def is_fresh_start(self):
        return (
            (self.ec2_ami_id is None)
            and (self.db_snapshot_id is None)
            and (self.ec2_inst_id is None)
            and (self.db_inst_id is None)
            and (self.is_ec2_ami_deleted is False)
            and (self.is_db_snapshot_deleted is False)
        )

    def is_ec2_ami_created(self) -> bool:
        return self.ec2_ami_id is not None

    def is_db_snapshot_created(self) -> bool:
        return self.db_snapshot_id is not None

    def is_ec2_instance_created(self) -> bool:
        return self.ec2_inst_id is not None

    def is_db_instance_created(self) -> bool:
        return self.db_inst_id is not None


@dataclasses.dataclass
class CreateUpdatedServerWorkflow(Workflow):
    db_snapshot_id: T.Optional[str] = dataclasses.field(default=None)
    ec2_inst_id: T.Optional[str] = dataclasses.field(default=None)
    db_inst_id: T.Optional[str] = dataclasses.field(default=None)
    is_ec2_ami_deleted: bool = dataclasses.field(default=False)
    is_db_snapshot_deleted: bool = dataclasses.field(default=False)

    def is_db_snapshot_created(self) -> bool:
        return self.db_snapshot_id is not None

    def is_ec2_instance_created(self) -> bool:
        return self.ec2_inst_id is not None

    def is_db_instance_created(self) -> bool:
        return self.db_inst_id is not None


@dataclasses.dataclass
class DeleteServerWorkflow(Workflow):
    ec2_ami_id: T.Optional[str] = dataclasses.field(default=None)
    db_snapshot_id: T.Optional[str] = dataclasses.field(default=None)
    is_ec2_ami_available: bool = dataclasses.field(default=False)
    is_db_snapshot_available: bool = dataclasses.field(default=False)
    is_ec2_deleted: bool = dataclasses.field(default=False)
    is_db_deleted: bool = dataclasses.field(default=False)

    def is_fresh_start(self):
        return (self.ec2_ami_id is None) and (self.db_snapshot_id is None)

    def is_ec2_ami_created(self) -> bool:
        return self.ec2_ami_id is not None

    def is_db_snapshot_created(self) -> bool:
        return self.db_snapshot_id is not None


class ServerWorkflowMixin:  # pragma: no cover
    """
    Server Workflow Mixin class that contains all the server workflow methods.
    """

    def ensure_server_is_ready_for_clone(
        self: "Server",
        bsm: "BotoSesManager",
    ) -> T.Tuple[simple_aws_ec2.Ec2Instance, simple_aws_rds.RDSDBInstance]:
        ec2_inst = self.ensure_rds_exists(bsm=bsm)
        if self.metadata.is_rds_running() is False:
            raise FailedToStartServerError(
                f"RDS DB instance {self.id!r} is not running, so we cannot create DB snapshot then clone the server!"
            )
        return ec2_inst, self.metadata.rds_inst

    @logger.emoji_block(
        msg="🧬🖥🛢Create Cloned Server",
        emoji="🧬",
    )
    def create_cloned_server(
        self: "Server",
        bsm: "BotoSesManager",
        workflow_id: str,
        s3path_workflow: S3Path,
        new_server_id: str,
        stack_exports: "StackExports",
        snapshot_id: T.Optional[str] = None,
        skip_reboot: bool = False,
        delete_ami_afterwards: bool = False,
        delete_snapshot_afterwards: bool = False,
    ):
        """
        Implement :ref:`create-cloned-server`.
        """
        aws_console = AWSConsole.from_bsm(bsm=bsm)
        workflow = CreateClonedServerWorkflow.load(
            bsm=bsm, s3_path=s3path_workflow, workflow_id=workflow_id
        )
        if snapshot_id is not None:
            workflow.db_snapshot_id = snapshot_id
            if delete_snapshot_afterwards is True:
                logger.info(
                    f"You provide the snapshot id explicitly, "
                    f"we cannot delete it afterward, "
                    f"set delete_snapshot_afterwards = False"
                )
                delete_snapshot_afterwards = False
        logger.info(f"Clone from {self.id!r} to {new_server_id!r}")

        logger.info("Check new server configurations ...")
        new_server = self.get(bsm=bsm, server_id=new_server_id)
        if new_server.config.is_ready_for_create_cloned_server() is False:
            raise ValueError("server config is not ready for create cloned server")
        logger.info("✅ new server configuration is fine.")

        # --- create AMI and DB Snapshot
        utc_now = get_utc_now()

        if workflow.is_ec2_ami_created() is False:
            with logger.nested():
                self.ensure_ec2_exists(bsm=bsm)
                ami_name = self._get_ec2_ami_name(utc_now=utc_now)
                res = self.create_ec2_ami(
                    bsm=bsm,
                    ami_name=ami_name,
                    skip_reboot=skip_reboot,
                    wait=False,
                )
                ami_id = res["ImageId"]
            url = aws_console.ec2.get_ami(ami_id)
            logger.info(f"🆕🖥📸Created EC2 AMI {ami_id!r}, preview at {url}")
            workflow.ec2_ami_id = ami_id
            workflow.dump(bsm=bsm, s3_path=s3path_workflow)
        else:
            logger.info("Skip creating EC2 AMI, AMI is already created.")

        if workflow.is_db_snapshot_created() is False:
            with logger.nested():
                # todo, if the RDS is not running, run it, then create snapshot, then stop it at the end
                snapshot_id = self._get_db_snapshot_id(utc_now=utc_now)
                self.ensure_rds_exists(bsm=bsm)
                res = self.create_db_snapshot(
                    bsm=bsm, snapshot_id=snapshot_id, wait=False
                )
                snapshot_id = res["DBSnapshot"]["DBSnapshotIdentifier"]
            url = aws_console.rds.get_snapshot(snapshot_id)
            logger.info(f"🆕🛢📸Created DB Snapshot {snapshot_id!r}, preview at {url}")
            workflow.db_snapshot_id = snapshot_id
            workflow.dump(bsm=bsm, s3_path=s3path_workflow)
        else:
            logger.info("Skip creating DB snapshot, snapshot is already created.")

        # --- create RDS
        if workflow.is_db_instance_created() is False:
            logger.info("wait for DB Snapshot to be available ...")
            snapshot = simple_aws_rds.RDSDBSnapshot(
                db_snapshot_identifier=workflow.db_snapshot_id
            )
            snapshot.wait_for_available(rds_client=bsm.rds_client)

            with logger.nested():
                new_server.ensure_rds_not_exists(bsm=bsm)
                res = new_server.create_rds_from_snapshot(
                    bsm=bsm,
                    stack_exports=stack_exports,
                    db_snapshot_id=workflow.db_snapshot_id,
                    check=True,
                    wait=True,
                )
                db_inst_id = res["DBInstance"]["DBInstanceIdentifier"]

            url = aws_console.rds.get_database_instance(db_inst_id)
            logger.info(f"🆕🛢Created DB Instance {db_inst_id!r}, preview at {url}")
            workflow.db_inst_id = db_inst_id
            workflow.dump(bsm=bsm, s3_path=s3path_workflow)
        else:
            logger.info("Skip creating DB instance, DB is already created.")

        # --- create EC2
        if workflow.is_ec2_instance_created() is False:
            logger.info("wait for EC2 AMI to be available ...")
            image = simple_aws_ec2.Image(id=workflow.ec2_ami_id)
            image.wait_for_available(ec2_client=bsm.ec2_client)

            with logger.nested():
                new_server.ensure_ec2_not_exists(bsm=bsm)
                res = new_server.create_ec2(
                    bsm=bsm,
                    stack_exports=stack_exports,
                    ami_id=workflow.ec2_ami_id,
                    check=True,
                    wait=True,
                )
                ec2_inst_id = res["Instances"][0]["InstanceId"]
            url = aws_console.ec2.get_instance(ec2_inst_id)
            logger.info(f"🆕🖥Created EC2 instance {ec2_inst_id!r}, preview at {url}")
            workflow.ec2_inst_id = ec2_inst_id
            workflow.dump(bsm=bsm, s3_path=s3path_workflow)
        else:
            logger.info("Skip creating EC2 instance, EC2 is already created.")

        if delete_ami_afterwards:
            if workflow.is_ec2_ami_deleted is False:
                logger.info("Delete AMI ...")
                image = simple_aws_ec2.Image(id=workflow.ec2_ami_id)
                image.deregister(
                    ec2_client=bsm.ec2_client,
                    delete_snapshot=True,  # also delete the snapshot
                    skip_prompt=True,
                )
                logger.info("✅Done")
                workflow.is_ec2_ami_deleted = True
                workflow.dump(bsm=bsm, s3_path=s3path_workflow)
            else:
                logger.info("Skip delete EC2 AMI, AMI is already deleted.")
        else:
            logger.info("We don't delete temp EC2 AMI afterward.")

        if delete_snapshot_afterwards:
            if workflow.is_db_snapshot_deleted is False:
                logger.info("Delete Snapshot ...")
                bsm.rds_client.delete_db_snapshot(
                    DBSnapshotIdentifier=workflow.db_snapshot_id,
                )
                logger.info("✅Done")
                workflow.is_db_snapshot_deleted = True
                workflow.dump(bsm=bsm, s3_path=s3path_workflow)
            else:
                logger.info("Skip delete DB snapshot, snapshot is already deleted.")
        else:
            logger.info("We don't delete temp DB snapshot afterward.")

    @logger.emoji_block(
        msg="🆕🖥🛢Create Updated Server",
        emoji="🆕",
    )
    def create_updated_server(
        self: "Server",
        bsm: "BotoSesManager",
        workflow_id: str,
        s3path_workflow: S3Path,
        new_server_id: str,
        ami_id: str,
        stack_exports: "StackExports",
        snapshot_id: T.Optional[str] = None,
        delete_snapshot_afterwards: bool = False,
    ):
        """
        Implement :ref:`create-updated-server`.

        :param ami_id: the AMI ID to create the new EC2 instance.
        :param snapshot_id: if None, it will try to run the db instance and wait
            it become available, then create a snapshot from it; if None and
            the db instant not exists, then raise error immediately; if provided
            use the given snapshot_id to create the db instance.
        """
        aws_console = AWSConsole.from_bsm(bsm=bsm)
        workflow = CreateUpdatedServerWorkflow.load(
            bsm=bsm,
            s3_path=s3path_workflow,
            workflow_id=workflow_id,
        )
        if snapshot_id is not None:
            workflow.db_snapshot_id = snapshot_id
            if delete_snapshot_afterwards is True:
                logger.info(
                    f"You provide the snapshot id explicitly, "
                    f"we cannot delete it afterward, "
                    f"set delete_snapshot_afterwards = False"
                )
                delete_snapshot_afterwards = False
        logger.info(f"Create updated server from {self.id!r} to {new_server_id!r}")

        logger.info("Check new server configurations ...")
        new_server = self.get(bsm=bsm, server_id=new_server_id)
        if new_server.config.is_ready_for_create_cloned_server() is False:
            raise ValueError("server config is not ready for create cloned server")
        logger.info("✅ new server configuration is fine.")

        # --- create DB Snapshot
        utc_now = get_utc_now()
        if workflow.is_db_snapshot_created() is False:
            with logger.nested():
                snapshot_id = self._get_db_snapshot_id(utc_now=utc_now)
                # todo: if the RDS is not running, run it, then create snapshot, then stop it at the end
                self.ensure_rds_exists(bsm=bsm)
                res = self.create_db_snapshot(
                    bsm=bsm, snapshot_id=snapshot_id, wait=False
                )
                snapshot_id = res["DBSnapshot"]["DBSnapshotIdentifier"]
            url = aws_console.rds.get_snapshot(snapshot_id)
            logger.info(f"🆕🛢📸Created DB Snapshot {snapshot_id!r}, preview at {url}")
            workflow.db_snapshot_id = snapshot_id
            workflow.dump(bsm=bsm, s3_path=s3path_workflow)
        else:
            logger.info("Skip creating DB snapshot, snapshot is already created.")

        # --- create RDS
        if workflow.is_db_instance_created() is False:
            logger.info("wait for DB Snapshot to be available ...")
            snapshot = simple_aws_rds.RDSDBSnapshot(
                db_snapshot_identifier=workflow.db_snapshot_id
            )
            snapshot.wait_for_available(rds_client=bsm.rds_client)

            with logger.nested():
                new_server.ensure_rds_not_exists(bsm=bsm)
                res = new_server.create_rds_from_snapshot(
                    bsm=bsm,
                    stack_exports=stack_exports,
                    db_snapshot_id=workflow.db_snapshot_id,
                    check=True,
                    wait=True,
                )
                db_inst_id = res["DBInstance"]["DBInstanceIdentifier"]

            url = aws_console.rds.get_database_instance(db_inst_id)
            logger.info(f"🆕🛢Created DB Instance {db_inst_id!r}, preview at {url}")
            workflow.db_inst_id = db_inst_id
            workflow.dump(bsm=bsm, s3_path=s3path_workflow)
        else:
            logger.info("Skip creating DB instance, DB is already created.")

        # --- create EC2
        if workflow.is_ec2_instance_created() is False:
            with logger.nested():
                new_server.ensure_ec2_not_exists(bsm=bsm)
                res = new_server.create_ec2(
                    bsm=bsm,
                    stack_exports=stack_exports,
                    ami_id=ami_id,
                    check=True,
                    wait=True,
                )
                ec2_inst_id = res["Instances"][0]["InstanceId"]
            url = aws_console.ec2.get_instance(ec2_inst_id)
            logger.info(f"🆕🖥Created EC2 instance {ec2_inst_id!r}, preview at {url}")
            workflow.ec2_inst_id = ec2_inst_id
            workflow.dump(bsm=bsm, s3_path=s3path_workflow)
        else:
            logger.info("Skip creating EC2 instance, EC2 is already created.")

        if delete_snapshot_afterwards:
            if workflow.is_db_snapshot_deleted is False:
                logger.info("Delete Snapshot ...")
                bsm.rds_client.delete_db_snapshot(
                    DBSnapshotIdentifier=workflow.db_snapshot_id,
                )
                logger.info("✅Done")
                workflow.is_db_snapshot_deleted = True
                workflow.dump(bsm=bsm, s3_path=s3path_workflow)
            else:
                logger.info("Skip delete DB snapshot, snapshot is already deleted.")
        else:
            logger.info("We don't delete temp DB snapshot afterward.")

    @logger.emoji_block(
        msg="🗑🖥🛢Delete Server",
        emoji="🗑",
    )
    def delete_server(
        self: "Server",
        bsm: "BotoSesManager",
        workflow_id: str,
        s3path_workflow: S3Path,
        skip_reboot: bool = False,
        create_backup_ec2_ami: bool = True,
        create_backup_db_snapshot: bool = True,
        skip_prompt: bool = False,
    ):
        """
        Implement :ref:`delete-server`.
        """
        aws_console = AWSConsole.from_bsm(bsm=bsm)
        workflow = DeleteServerWorkflow.load(
            bsm=bsm, s3_path=s3path_workflow, workflow_id=workflow_id
        )
        logger.info(f"Delete {self.id!r}")
        if skip_prompt is False:
            prompt_for_confirm(
                msg=(f"💥Are you sure you want to DELETE server {self.id!r}?")
            )

        # --- create AMI and DB Snapshot
        utc_now = get_utc_now()

        if create_backup_ec2_ami:
            if workflow.is_ec2_ami_created() is False:
                with logger.nested():
                    ami_name = self._get_ec2_ami_name()
                    ami_name = f"{ami_name}-final-backup"
                    res = self.create_ec2_ami(
                        bsm=bsm,
                        ami_name=ami_name,
                        skip_reboot=skip_reboot,
                        wait=False,
                    )
                    ami_id = res["ImageId"]
                url = aws_console.ec2.get_ami(ami_id)
                logger.info(f"🗑🖥📸Created EC2 AMI {ami_id!r}, preview at {url}")
                workflow.ec2_ami_id = ami_id
                workflow.dump(bsm=bsm, s3_path=s3path_workflow)
            else:
                logger.info("Skip creating EC2 AMI, AMI is already created.")
        else:
            logger.info("We don't creating final EC2 AMI backup.")

        if create_backup_db_snapshot:
            if workflow.is_db_snapshot_created() is False:
                with logger.nested():
                    snapshot_id = self._get_db_snapshot_id(utc_now=utc_now)
                    snapshot_id = f"{snapshot_id}-final-backup"
                    res = self.create_db_snapshot(
                        bsm=bsm,
                        snapshot_id=snapshot_id,
                        wait=False,
                    )
                    snapshot_id = res["DBSnapshot"]["DBSnapshotIdentifier"]
                url = aws_console.rds.get_snapshot(snapshot_id)
                logger.info(
                    f"🗑🛢📸Created DB Snapshot {snapshot_id!r}, preview at {url}"
                )
                workflow.db_snapshot_id = snapshot_id
                workflow.dump(bsm=bsm, s3_path=s3path_workflow)
            else:
                logger.info("Skip creating DB snapshot, snapshot is already created.")
        else:
            logger.info("We don't creating final DB snapshot backup.")

        if create_backup_ec2_ami:
            if workflow.is_ec2_ami_available is False:
                logger.info("wait for EC2 AMI to be available ...")
                image = simple_aws_ec2.Image(id=workflow.ec2_ami_id)
                image.wait_for_available(ec2_client=bsm.ec2_client)
                workflow.is_ec2_ami_available = True
                workflow.dump(bsm=bsm, s3_path=s3path_workflow)

        if create_backup_db_snapshot:
            if workflow.is_db_snapshot_available is False:
                logger.info("wait for DB Snapshot to be available ...")
                snapshot = simple_aws_rds.RDSDBSnapshot(
                    db_snapshot_identifier=workflow.db_snapshot_id
                )
                snapshot.wait_for_available(rds_client=bsm.rds_client)
                workflow.is_db_snapshot_available = True
                workflow.dump(bsm=bsm, s3_path=s3path_workflow)

        # --- delete EC2
        if workflow.is_ec2_deleted is False:
            if self.metadata.is_ec2_exists():
                with logger.nested():
                    res = self.delete_ec2(
                        bsm=bsm,
                        check=False,
                    )
                    ec2_inst_id = self.metadata.ec2_inst.id
                    url = aws_console.ec2.get_instance(ec2_inst_id)
                logger.info(f"🗑🖥Delete EC2 instance {ec2_inst_id!r}, verify at {url}")
            else:
                logger.info("Skip terminate EC2 instance, EC2 is already terminated.")
            workflow.is_ec2_deleted = True
            workflow.dump(bsm=bsm, s3_path=s3path_workflow)
        else:
            logger.info("Skip terminate EC2 instance, EC2 is already terminated.")

        # --- delete RDS
        if workflow.is_db_deleted is False:
            if self.metadata.is_rds_exists():
                with logger.nested():
                    res = self.delete_rds(
                        bsm=bsm,
                        create_final_snapshot=False,
                        check=False,
                    )
                    db_inst_id = self.metadata.rds_inst.id
                    url = aws_console.rds.get_database_instance(db_inst_id)
                logger.info(f"🗑🛢Delete DB Instance {db_inst_id!r}, verify at {url}")
            else:
                logger.info("Skip delete DB instance, DB is already deleted.")
            workflow.is_db_deleted = True
            workflow.dump(bsm=bsm, s3_path=s3path_workflow)
        else:
            logger.info("Skip delete DB instance, DB is already deleted.")



# -*- coding: utf-8 -*-

"""
todo: doc string
"""

import typing as T
from pathlib import Path
from datetime import datetime, timezone

from boto_session_manager import BotoSesManager
from aws_ssm_run_command.api import better_boto as ssm_better_boto
import simple_aws_ec2.api as simple_aws_ec2
import simple_aws_rds.api as simple_aws_rds

from acore_paths.api import path_acore_server_bootstrap_cli
from acore_constants.api import TagKey
from acore_server_config.api import EnvEnum
from acore_db_ssh_tunnel import api as acore_db_ssh_tunnel
from .vendor.hashes import hashes

from .constants import EC2_USERNAME, DB_PORT
from .wserver_infra_exports import StackExports
from .logger import logger
from .utils import get_utc_now


if T.TYPE_CHECKING:  # pragma: no cover
    from .server import Server
    from mypy_boto3_ec2.type_defs import (
        CreateImageResultTypeDef,
        ReservationResponseTypeDef,
        StartInstancesResultTypeDef,
        StopInstancesResultTypeDef,
        TerminateInstancesResultTypeDef,
    )
    from mypy_boto3_rds.type_defs import (
        CreateDBSnapshotResultTypeDef,
        CreateDBInstanceResultTypeDef,
        RestoreDBInstanceFromDBSnapshotResultTypeDef,
        StartDBInstanceResultTypeDef,
        StopDBInstanceResultTypeDef,
        DeleteDBInstanceResultTypeDef,
    )


class ServerOperationMixin:  # pragma: no cover
    """
    Server Operation Mixin class that contains all the server operation methods.
    """

    def _get_ec2_ami_name(
        self: "Server",
        utc_now: T.Optional[datetime] = None,
    ) -> str:
        """
        Get the EC2 AMI id for this server, the snapshot id
        naming convention is "${server_id}-%Y-%m-%d-%H-%M-%S".
        """
        if utc_now is None:
            utc_now = get_utc_now()
        snapshot_id = "{}-{}".format(
            self.id,
            utc_now.strftime("%Y-%m-%d-%H-%M-%S"),
        )
        return snapshot_id

    def _get_db_snapshot_id(
        self: "Server",
        utc_now: T.Optional[datetime] = None,
    ) -> str:
        """
        Get the db snapshot id for this server, the snapshot id
        naming convention is "${server_id}-%Y-%m-%d-%H-%M-%S".
        """
        if utc_now is None:
            utc_now = get_utc_now()
        snapshot_id = "{}-{}".format(
            self.id,
            utc_now.strftime("%Y-%m-%d-%H-%M-%S"),
        )
        return snapshot_id

    def build_bootstrap_command(
        self: "Server",
        python_version: str = "3.8",
        acore_soap_app_version: T.Optional[str] = None,
        acore_db_app_version: T.Optional[str] = None,
        acore_server_bootstrap_version: T.Optional[str] = None,
    ) -> str:
        """
        构建需要在 EC2 服务器上运行的 bootstrap 命令.
        """
        script_url = "https://raw.githubusercontent.com/MacHu-GWU/acore_server_bootstrap-project/main/install.py"
        main_part = (
            f"sudo /home/ubuntu/.pyenv/shims/python{python_version} -c "
            f'"$(curl -fsSL {script_url})"'
        )
        parts = [main_part]
        if acore_soap_app_version:
            parts.append(
                f"--acore_soap_app_version {self.config.acore_soap_app_version}"
            )
        if acore_db_app_version:
            parts.append(f"--acore_db_app_version {self.config.acore_db_app_version}")
        if acore_server_bootstrap_version:
            parts.append(
                f"--acore_server_bootstrap_version {self.config.acore_server_bootstrap_version}"
            )
        cmd = " ".join(parts)
        return cmd

    @logger.emoji_block(
        msg="🆕🖥Create EC2 Instance",
        emoji="🖥",
    )
    def create_ec2(
        self: "Server",
        bsm: "BotoSesManager",
        stack_exports: "StackExports",
        ami_id: T.Optional[str] = None,
        instance_type: T.Optional[str] = None,
        python_version: str = "3.8",
        acore_soap_app_version: T.Optional[str] = None,
        acore_db_app_version: T.Optional[str] = None,
        acore_server_bootstrap_version: T.Optional[str] = None,
        tags: T.Optional[T.Dict[str, str]] = None,
        check: bool = True,
        wait: bool = True,
        **kwargs,
    ) -> "ReservationResponseTypeDef":
        """
        为服务器创建一台新的 EC2. 注意, 一般先创建 RDS, 等 RDS 已经在运行了, 再创建 EC2.
        因为 EC2 有一个 bootstrap 的过程, 这个过程中需要跟数据库通信.

        :param ami_id: 你要从哪个 AMI 来创建 EC2? 如果不指定, 则使用 config 中的 AMI ID.
            这个参数之所以是可选是因为在有的 workflow 中, 我们已经知道 AMI ID 了;
            而有的 workflow 中, 我们要临时创建一个新的 AMI ID, 此时还不知道这个 ID.
        """
        bootstrap_command = self.build_bootstrap_command(
            python_version=python_version,
            acore_soap_app_version=acore_soap_app_version,
            acore_db_app_version=acore_db_app_version,
            acore_server_bootstrap_version=acore_server_bootstrap_version,
        )
        user_data_lines = [
            "#!/bin/bash",
            bootstrap_command,
        ]
        if check:
            self.ensure_ec2_not_exists(bsm=bsm)
        if tags is None:
            tags = dict()
        tags["Name"] = self.id
        tags[TagKey.SERVER_ID] = self.id  # the realm tag indicator has to match
        tags["tech:machine_creator"] = "acore_server_metadata"
        if ami_id is None:
            ami_id = self.config.ec2_ami_id
        if instance_type is None:
            instance_type = self.config.ec2_instance_type
        run_instances_kwargs = dict(
            ImageId=ami_id,
            InstanceType=instance_type,
            # only launch one instance for each realm
            MinCount=1,
            MaxCount=1,
            KeyName=self.config.ec2_key_name,
            SecurityGroupIds=[
                stack_exports.get_default_sg_id(server_id=self.id),
                stack_exports.get_ec2_sg_id(server_id=self.id),
                stack_exports.get_ssh_sg_id(),
            ],
            SubnetId=self.config.ec2_subnet_id,
            IamInstanceProfile=dict(Arn=stack_exports.get_ec2_instance_profile_arn()),
            TagSpecifications=[
                dict(
                    ResourceType="instance",
                    Tags=[dict(Key=k, Value=v) for k, v in tags.items()],
                ),
            ],
            **kwargs,
        )
        if any(
            [
                acore_soap_app_version,
                acore_db_app_version,
                acore_server_bootstrap_version,
            ]
        ):
            kwargs["UserData"] = "\n".join(user_data_lines)
        res = bsm.ec2_client.run_instances(**run_instances_kwargs)
        if wait:
            inst_id = res["Instances"][0]["InstanceId"]
            ec2_inst = simple_aws_ec2.Ec2Instance(id=inst_id, status="na")
            ec2_inst.wait_for_running(ec2_client=bsm.ec2_client, timeout=300)
        return res

    @logger.emoji_block(
        msg="🆕🛢Create RDS from scratch",
        emoji="🛢",
    )
    def create_rds_from_scratch(
        self: "Server",
        bsm: "BotoSesManager",
        stack_exports: "StackExports",
        db_instance_class: T.Optional[str] = None,
        engine_version: T.Optional[str] = None,
        multi_az: T.Optional[bool] = None,
        tags: T.Optional[T.Dict[str, str]] = None,
        check: bool = True,
        wait: bool = True,
        **kwargs,
    ) -> "CreateDBInstanceResultTypeDef":
        """
        不使用 DB Snapshot, 创建一台全新的数据库.

        - https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/rds/client/create_db_instance.html
        """
        if check:
            self.ensure_rds_not_exists(bsm=bsm)
        if tags is None:
            tags = dict()
        tags[TagKey.SERVER_ID] = self.id
        tags["tech:machine_creator"] = "acore_server"
        hashes.use_sha256()
        digest = hashes.of_str(self.config.db_admin_password, hexdigest=True)
        tags["tech:master_password_digest"] = digest
        if db_instance_class is None:
            db_instance_class = self.config.db_instance_class
        if engine_version is None:
            engine_version = self.config.db_engine_version
        if multi_az is None:
            multi_az = False
        res = bsm.rds_client.create_db_instance(
            DBInstanceIdentifier=self.id,
            DBInstanceClass=db_instance_class,
            Engine="mysql",
            EngineVersion=engine_version,
            MasterUsername=self.config.db_admin_username,
            MasterUserPassword=self.config.db_admin_password,
            MultiAZ=multi_az,
            DBSubnetGroupName=stack_exports.get_db_subnet_group_name(),
            PubliclyAccessible=False,  # you should never expose your database to the public
            AutoMinorVersionUpgrade=False,  # don't update MySQL minor version, PLEASE!
            VpcSecurityGroupIds=[
                stack_exports.get_default_sg_id(server_id=self.id),
            ],
            CopyTagsToSnapshot=True,
            Tags=[dict(Key=k, Value=v) for k, v in tags.items()],
            **kwargs,
        )
        if wait:
            rds_inst = simple_aws_rds.RDSDBInstance(id=self.id, status="na")
            rds_inst.wait_for_available(rds_client=bsm.rds_client, timeout=600)
        return res

    @logger.emoji_block(
        msg="🆕🛢Create RDS from snapshot",
        emoji="🛢",
    )
    def create_rds_from_snapshot(
        self: "Server",
        bsm: "BotoSesManager",
        stack_exports: "StackExports",
        db_snapshot_id: T.Optional[str] = None,
        db_instance_class: T.Optional[str] = None,
        multi_az: T.Optional[bool] = None,
        tags: T.Optional[T.Dict[str, str]] = None,
        check: bool = True,
        wait: bool = True,
        **kwargs,
    ) -> "RestoreDBInstanceFromDBSnapshotResultTypeDef":
        """
        为服务器创建一台新的数据库.

        :param bsm:
        :param stack_exports:
        :param db_snapshot_id: 要从哪个 DB Snapshot 来创建 RDS? 如果不指定, 则使用 config 中的 DB Snapshot ID.
            这个参数之所以是可选是因为在有的 workflow 中, 我们已经知道 DB Snapshot ID 了;
            而有的 workflow 中, 我们要临时创建一个新的 DB Snapshot ID, 此时还不知道这个 ID.
        :param db_instance_class:
        :param multi_az:
        :param tags:
        :param check:
        """
        if check:
            self.ensure_rds_not_exists(bsm=bsm)
        if tags is None:
            tags = dict()
        tags[TagKey.SERVER_ID] = self.id
        tags["tech:machine_creator"] = "acore_server"
        if db_snapshot_id is None:
            db_snapshot_id = self.config.db_snapshot_id
        if db_instance_class is None:
            db_instance_class = self.config.db_instance_class
        if multi_az is None:
            multi_az = False
        # locate master password digest from snapshot tags
        res = bsm.rds_client.describe_db_snapshots(
            DBSnapshotIdentifier=db_snapshot_id,
        )
        db_snapshot_list = res.get("DBSnapshots", [])
        if len(db_snapshot_list):
            db_snapshot_tags = {
                dct["Key"]: dct["Value"]
                for dct in db_snapshot_list[0].get("TagList", [])
            }
            master_password_digest = db_snapshot_tags.get("tech:master_password_digest")
            if master_password_digest:
                tags["tech:master_password_digest"] = master_password_digest
        res = bsm.rds_client.restore_db_instance_from_db_snapshot(
            DBInstanceIdentifier=self.id,
            DBSnapshotIdentifier=db_snapshot_id,
            DBInstanceClass=db_instance_class,
            MultiAZ=multi_az,
            DBSubnetGroupName=stack_exports.get_db_subnet_group_name(),
            PubliclyAccessible=False,  # you should never expose your database to the public
            AutoMinorVersionUpgrade=False,  # don't update MySQL minor version, PLEASE!
            VpcSecurityGroupIds=[
                stack_exports.get_default_sg_id(server_id=self.id),
            ],
            CopyTagsToSnapshot=True,
            Tags=[dict(Key=k, Value=v) for k, v in tags.items()],
            **kwargs,
        )
        if wait:
            rds_inst = simple_aws_rds.RDSDBInstance(id=self.id, status="na")
            rds_inst.wait_for_available(rds_client=bsm.rds_client, timeout=600)
        return res

    @logger.emoji_block(
        msg="🟢🖥Start EC2 instance",
        emoji="🖥",
    )
    def start_ec2(
        self: "Server",
        bsm: "BotoSesManager",
        check: bool = True,
        wait: bool = True,
    ) -> "StartInstancesResultTypeDef":
        """
        启动关闭着的 EC2.
        """
        if check:
            self.ensure_ec2_is_ready_to_start(bsm=bsm)
        ec2_inst = self.metadata.ec2_inst
        res = ec2_inst.start_instance(ec2_client=bsm.ec2_client)
        if wait:
            ec2_inst.wait_for_running(ec2_client=bsm.ec2_client, timeout=300)
        return res

    @logger.emoji_block(
        msg="🟢🛢Start RDS instance",
        emoji="🛢",
    )
    def start_rds(
        self: "Server",
        bsm: "BotoSesManager",
        check: bool = True,
        wait: bool = True,
    ) -> "StartDBInstanceResultTypeDef":
        """
        启动关闭着的 RDS.
        """
        if check:
            self.ensure_rds_is_ready_to_start(bsm=bsm)
        rds_inst = self.metadata.rds_inst
        res = rds_inst.start_db_instance(rds_client=bsm.rds_client)
        if wait:
            rds_inst.wait_for_available(rds_client=bsm.rds_client, timeout=600)
        return res

    @logger.emoji_block(
        msg="Associate EIP Address",
        emoji="🖥",
    )
    def associate_eip_address(
        self: "Server",
        bsm: "BotoSesManager",
        allow_reassociation: bool = False,
        check: bool = True,
    ) -> T.Optional[dict]:
        """
        给 EC2 关联 EIP.
        """
        if self.config.ec2_eip_allocation_id is not None:
            if check:
                ec2_inst = self.ensure_ec2_exists(bsm=bsm)
            else:
                ec2_inst = self.metadata.ec2_inst
            # check if this allocation id is already associated with an instance
            res = bsm.ec2_client.describe_addresses(
                AllocationIds=[self.config.ec2_eip_allocation_id]
            )
            address_data = res["Addresses"][0]
            instance_id = address_data.get("InstanceId", "invalid-instance-id")
            if instance_id == ec2_inst.id:  # already associated
                return None

            # associate eip address
            return bsm.ec2_client.associate_address(
                AllocationId=self.config.ec2_eip_allocation_id,
                InstanceId=ec2_inst.id,
                AllowReassociation=allow_reassociation,
            )
        return None

    @logger.emoji_block(
        msg="Update db master password",
        emoji="🛢",
    )
    def update_db_master_password(
        self: "Server",
        bsm: "BotoSesManager",
        check: bool = True,
    ) -> T.Optional[dict]:
        """
        更新数据库的 master password.
        """
        if check:
            rds_inst = self.ensure_rds_exists(bsm=bsm)
        else:
            rds_inst = self.metadata.rds_inst

        master_password = self.config.db_admin_password
        hashes.use_sha256()
        master_password_digest = hashes.of_str(master_password, hexdigest=True)
        if (
            rds_inst.tags.get("tech:master_password_digest", "invalid")
            == master_password_digest
        ):
            # do nothing
            return None

        response = bsm.rds_client.modify_db_instance(
            DBInstanceIdentifier=rds_inst.id,
            MasterUserPassword=master_password,
            ApplyImmediately=True,
        )

        bsm.rds_client.add_tags_to_resource(
            ResourceName=rds_inst.db_instance_arn,
            Tags=[
                dict(Key="tech:master_password_digest", Value=master_password_digest)
            ],
        )

        return response

    @logger.emoji_block(
        msg="🔴🖥Stop EC2 instance",
        emoji="🖥",
    )
    def stop_ec2(
        self: "Server",
        bsm: "BotoSesManager",
        check: bool = True,
        wait: bool = True,
    ) -> "StopInstancesResultTypeDef":
        """
        关闭 EC2.
        """
        if check:
            self.ensure_ec2_is_ready_to_stop(bsm=bsm)
        ec2_inst = self.metadata.ec2_inst
        res = ec2_inst.stop_instance(ec2_client=bsm.ec2_client)
        if wait:
            ec2_inst.wait_for_stopped(ec2_client=bsm.ec2_client, timeout=300)
        return res

    @logger.emoji_block(
        msg="🔴🛢Stop RDS instance",
        emoji="🛢",
    )
    def stop_rds(
        self: "Server",
        bsm: "BotoSesManager",
        check: bool = True,
        wait: bool = True,
    ) -> "StopDBInstanceResultTypeDef":
        """
        关闭 RDS.
        """
        if check:
            self.ensure_rds_is_ready_to_stop(bsm=bsm)
        rds_inst = self.metadata.rds_inst
        res = rds_inst.stop_db_instance(rds_client=bsm.rds_client)
        if wait:
            rds_inst.wait_for_stopped(rds_client=bsm.rds_client, timeout=300)
        return res

    @logger.emoji_block(
        msg="🗑🖥Stop EC2 instance",
        emoji="🖥",
    )
    def delete_ec2(
        self: "Server",
        bsm: "BotoSesManager",
        check: bool = True,
    ) -> "TerminateInstancesResultTypeDef":
        """
        删除 EC2.
        """
        if check:
            self.ensure_ec2_exists(bsm=bsm)
        ec2_inst = self.metadata.ec2_inst
        res = ec2_inst.terminate_instance(ec2_client=bsm.ec2_client)
        return res

    @logger.emoji_block(
        msg="🗑🛢Delete EC2 instance",
        emoji="🛢",
    )
    def delete_rds(
        self: "Server",
        bsm: "BotoSesManager",
        create_final_snapshot: T.Optional[bool] = None,
        check: bool = True,
    ) -> "DeleteDBInstanceResultTypeDef":
        """
        删除 RDS.
        """
        if create_final_snapshot is None:
            create_final_snapshot = self.env_name == EnvEnum.prd.value
        if check:
            self.ensure_rds_exists(bsm=bsm)
        rds_inst = self.metadata.rds_inst
        if create_final_snapshot:
            snapshot_id = self._get_db_snapshot_id()
            res = rds_inst.delete_db_instance(
                rds_client=bsm.rds_client,
                skip_final_snapshot=False,
                final_db_snapshot_identifier=snapshot_id,
                delete_automated_backups=False,
            )
        else:
            res = rds_inst.delete_db_instance(
                rds_client=bsm.rds_client,
                skip_final_snapshot=True,
                delete_automated_backups=True,
            )
        return res

    @logger.emoji_block(
        msg="🗑🛢Delete EC2 instance",
        emoji="🛢",
    )
    def bootstrap(
        self: "Server",
        bsm: "BotoSesManager",
        python_version: str = "3.8",
        acore_soap_app_version: T.Optional[str] = None,
        acore_db_app_version: T.Optional[str] = None,
        acore_server_bootstrap_version: T.Optional[str] = None,
    ):
        """
        远程执行 EC2 引导程序.

        这个命令不会失败. 它只是一个 async API call.
        """
        bootstrap_command = self.build_bootstrap_command(
            python_version=python_version,
            acore_soap_app_version=acore_soap_app_version,
            acore_db_app_version=acore_db_app_version,
            acore_server_bootstrap_version=acore_server_bootstrap_version,
        )
        return ssm_better_boto.run_shell_script_async(
            ssm_client=bsm.ssm_client,
            commands=bootstrap_command,
            instance_ids=self.metadata.ec2_inst.id,
        )

    @logger.emoji_block(
        msg="Run Check Server Status Cron Job",
        emoji="🖥",
    )
    def run_check_server_status_cron_job(
        self: "Server",
        bsm: "BotoSesManager",
    ):
        """
        启动检测游戏服务器状态的定时任务.

        这个命令不会失败. 它只是一个 async API call.
        """
        return ssm_better_boto.run_shell_script_async(
            ssm_client=bsm.ssm_client,
            commands=(
                f"sudo -H -u ubuntu "
                f"{path_acore_server_bootstrap_cli} run_check_server_status_cron_job"
            ),
            instance_ids=self.metadata.ec2_inst.id,
        )

    @logger.emoji_block(
        msg="Stop Check Server Status Cron Job",
        emoji="🖥",
    )
    def stop_check_server_status_cron_job(
        self: "Server",
        bsm: "BotoSesManager",
    ):
        """
        停止检测游戏服务器状态的定时任务.

        这个命令不会失败. 它只是一个 async API call.
        """
        return ssm_better_boto.run_shell_script_async(
            ssm_client=bsm.ssm_client,
            commands=(
                "sudo -H -u ubuntu "
                f"{path_acore_server_bootstrap_cli} stop_check_server_status_cron_job"
            ),
            instance_ids=self.metadata.ec2_inst.id,
        )

    @logger.emoji_block(
        msg="Run Game Server",
        emoji="🖥",
    )
    def run_game_server(
        self: "Server",
        bsm: "BotoSesManager",
    ):
        """
        启动魔兽世界游戏服务器.

        这个命令不会失败. 它只是一个 async API call.
        """
        return ssm_better_boto.run_shell_script_async(
            ssm_client=bsm.ssm_client,
            commands=(
                f"sudo -H -u ubuntu " f"{path_acore_server_bootstrap_cli} run_server"
            ),
            instance_ids=self.metadata.ec2_inst.id,
        )

    @logger.emoji_block(
        msg="Stop Game Server",
        emoji="🖥",
    )
    def stop_game_server(
        self: "Server",
        bsm: "BotoSesManager",
    ):
        """
        停止魔兽世界游戏服务器.

        这个命令不会失败. 它只是一个 async API call.
        """
        return ssm_better_boto.run_shell_script_async(
            ssm_client=bsm.ssm_client,
            commands=(
                f"sudo -H -u ubuntu " f"{path_acore_server_bootstrap_cli} stop_server"
            ),
            instance_ids=self.metadata.ec2_inst.id,
        )

    @property
    def wow_status(self: "Server") -> str:
        """
        从 EC2 的 Tag 中获取魔兽世界服务器的状态.

        1. 如果 EC2 或 RDS 任意一个不存在或是已被删除 则返回 "deleted".
        2. 如果 EC2 或 RDS 不在线, 则返回 ``stopped``.
        3. 如果 EC2 或 RDS 都在线, tag 中不存在测量数据, 则返回 "stopped"
        4. 如果 EC2 或 RDS 都在线, tag 中有测量数据且没有过期, 则返回 tag 中的数据, 值可能是
            "123 players" (数字是在线人数), 或是 "stopped"
        5. 如果 EC2 或 RDS 都在线, tag 中有测量数据且过期了, 则返回 "stopped"
        """
        if self.metadata.is_exists() is False:
            return "deleted"
        elif self.metadata.is_running() is False:
            return "stopped"
        elif TagKey.WOW_STATUS not in self.metadata.ec2_inst.tags:
            return "stopped"
        else:
            wow_status = self.metadata.ec2_inst.tags[TagKey.WOW_STATUS]
            measure_time = self.metadata.ec2_inst.tags[TagKey.WOW_STATUS_MEASURE_TIME]
            measure_time = datetime.fromisoformat(measure_time)
            utc_now = datetime.utcnow().replace(tzinfo=timezone.utc)
            elapsed = (utc_now - measure_time).total_seconds()
            if elapsed > 300:
                return "stopped"
            else:
                return wow_status

    def _get_path_pem_file(self: "Server", bsm: "BotoSesManager") -> Path:
        path = (
            Path.home()
            .joinpath("ec2-pem")
            .joinpath(
                bsm.aws_account_alias,
                bsm.aws_region,
                f"{self.metadata.ec2_inst.key_name}.pem",
            )
        )
        if path.exists() is False:
            raise FileNotFoundError(
                f"Failed to locate the pem file: {path}, "
                "make sure you have the pem file at ${HOME}/ec2-pem/${AWS_ACCOUNT_ALIAS}/${AWS_REGION}/${KEY_NAME}.pem"
            )
        return path

    def create_ssh_tunnel(
        self: "Server",
        bsm: "BotoSesManager",
        path_pem_file: T.Optional[Path] = None,
    ):
        """
        创建一个本地的 SSH Tunnel, 用于本地数据库开发.
        """
        if self.metadata.is_running() is False:
            raise ConnectionError(
                "cannot create ssh tunnel, EC2 or RDS is not running."
            )
        if path_pem_file is None:
            path_pem_file = self._get_path_pem_file(bsm=bsm)
        acore_db_ssh_tunnel.create_ssh_tunnel(
            path_pem_file=path_pem_file,
            db_host=self.metadata.rds_inst.endpoint,
            db_port=DB_PORT,
            jump_host_username=EC2_USERNAME,
            jump_host_public_ip=self.metadata.ec2_inst.public_ip,
        )

    def list_ssh_tunnel(
        self: "Server",
        bsm: "BotoSesManager",
        path_pem_file: "Path" = None,
    ):
        """
        列出所有正在运行中的 SSH Tunnel.
        """
        if path_pem_file is None:
            path_pem_file = self._get_path_pem_file(bsm=bsm)
        acore_db_ssh_tunnel.list_ssh_tunnel(path_pem_file)

    def test_ssh_tunnel(self: "Server"):
        """
        通过运行一个简单的 SQL 语句来测试 SSH Tunnel 是否正常工作.
        """
        if self.metadata.is_running() is False:
            raise ConnectionError("cannot test ssh tunnel, EC2 or RDS is not running.")
        acore_db_ssh_tunnel.test_ssh_tunnel(
            db_port=DB_PORT,
            db_username=self.config.db_username,
            db_password=self.config.db_password,
            db_name="acore_auth",
        )

    def kill_ssh_tunnel(
        self: "Server",
        bsm: "BotoSesManager",
        path_pem_file: "Path" = None,
    ):
        """
        关闭所有正在运行中的 SSH Tunnel.
        """
        if path_pem_file is None:
            path_pem_file = self._get_path_pem_file(bsm=bsm)
        acore_db_ssh_tunnel.kill_ssh_tunnel(path_pem_file)

    @logger.emoji_block(
        msg="🆕🖥📸Create new EC2 AMI",
        emoji="📸",
    )
    def create_ec2_ami(
        self: "Server",
        bsm: "BotoSesManager",
        ami_name: T.Optional[str] = None,
        utc_now: T.Optional[datetime] = None,
        skip_reboot: bool = True,
        check: bool = True,
        wait: bool = True,
    ) -> "CreateImageResultTypeDef":
        """
        Create a new AMI from existing EC2.
        """
        if ami_name is None:
            ami_name = self._get_ec2_ami_name(utc_now)
        if check:
            ec2_inst = self.ensure_ec2_exists(bsm=bsm)
        else:
            ec2_inst = self.metadata.ec2_inst
        logger.info(
            f"create image {ami_name!r} from ec2 instance {self.metadata.ec2_inst.id}"
        )
        res = bsm.ec2_client.create_image(
            InstanceId=self.metadata.ec2_inst.id,
            Name=ami_name,
            NoReboot=skip_reboot,
            TagSpecifications=[
                {
                    "ResourceType": "image",
                    "Tags": [{"Key": k, "Value": v} for k, v in ec2_inst.tags.items()],
                }
            ],
        )
        ami_id = res["ImageId"]
        if wait:
            image = simple_aws_ec2.Image(id=ami_id)
            image.wait_for_available(ec2_client=bsm.ec2_client)
        else:
            logger.info("skip waiting for available")
        return res

    @logger.emoji_block(
        msg="🆕🛢📸Create new DB Snapshot",
        emoji="📸",
    )
    def create_db_snapshot(
        self: "Server",
        bsm: "BotoSesManager",
        snapshot_id: T.Optional[str] = None,
        check: bool = True,
        wait: bool = True,
    ) -> "CreateDBSnapshotResultTypeDef":
        if snapshot_id is None:
            snapshot_id = self._get_db_snapshot_id()
        if check:
            rds_inst = self.ensure_rds_exists(bsm=bsm)
        else:
            rds_inst = self.metadata.rds_inst
        logger.info(
            f"create db snapshot {snapshot_id!r} from db instance {rds_inst.id}"
        )
        res = bsm.rds_client.create_db_snapshot(
            DBSnapshotIdentifier=snapshot_id,
            DBInstanceIdentifier=rds_inst.id,
        )
        if wait:
            snap = simple_aws_rds.RDSDBSnapshot(db_snapshot_identifier=snapshot_id)
            snap.wait_for_available(rds_client=bsm.rds_client)
        else:
            logger.info("skip waiting for available")
        return res
